/* 
 * find.js
 * Pedro Fonseca, 2012.
 */
var re;
var root;
var rawTerm;
var occurrences;
var queuedMatches;
var parentBeforeSelection;
var previousSiblingBeforeSelection;

/* This helper to determine the scroll offset was taken from: 
 * http://stackoverflow.com/questions/8922107/javascript-scrollintoview-middle-alignment 
 */
Element.prototype.documentOffsetTop = function () {
    return this.offsetTop + ( this.offsetParent ? this.offsetParent.documentOffsetTop() : 0 );
};


function resetQueue() {
    clearFindSelection();
    root = document.body.getElementsByTagName("pre")[0].firstChild;
    occurrences = 0;
    queuedMatches = Array();
    
    parentBeforeSelection = null;
    previousSiblingBeforeSelection = null;
}

function setTerm(t) {
    var clean = t.trim();
    var didChange = false;

    if (clean != rawTerm) {
        rawTerm = clean;
        
        /* If the term is a regular expression, remove the limiters because the 
         * RegExp class will be used. Otherwise, regular expression characters 
         * must be escaped.
         */
        if (clean.length > 2 && t[0] == "/" && clean[clean.length - 1] == "/") {
            clean = clean.substring(1, clean.length - 2);
        } else {
            clean = clean.replace(/[-[\]{}()*+?.,\\^$|#\s]/g, "\\$&");
        }

        re = new RegExp(clean, "gim");
        resetQueue();
        didChange = true; 
    }
    
    return didChange;
}

/* Clear the previous selection.
 *
 * Returns the normalized text node.
 */
function clearFindSelection() {   
    var oldSelection = document.getElementById("searchResult");
    if (oldSelection) {
        var replacement = document.createTextNode(oldSelection.textContent);
        oldSelection.parentNode.replaceChild(replacement, oldSelection);
        
        /* The normalization collapses adjacent text nodes into the first. */
        var originalAfterNormalized = replacement.previousSibling;
        replacement.parentNode.normalize();
       
        /* Edge case: the selected text is at the start of a text node. It
         * appears that when this happens a new node is created, thus the
         * originalAfterNormalized is empty and not in the DOM anymore.
         *
         * This is not an issue if the text is highlighted, because the document
         * structure generally prevents more than one match per text node. In
         * the case of keywords the text node is a single word. Comments that
         * span multiple lines are interleaved by line numbers, which are
         * wrapped in span tags, thus separating the text contents. Moreover,
         * the text immediately at the start of a comment, the limiter, is not
         * likely to appear repeated.
         *
         * String literals are the exception, if the search is for the string
         * limiter, as in most cases that limiter will begin and end the text
         * node. In that case the parent is the span tag that wraps the string.
         */
        var original = null;
        if (originalAfterNormalized.length > 0) {
            original = originalAfterNormalized;
        } else if (previousSiblingBeforeSelection !== null) {
            original = previousSiblingBeforeSelection.nextSibling;
        } else if (parentBeforeSelection !== null) {
            original = parentBeforeSelection.firstChild;
        }
        
        return original;
    }
    
    return null;
}

/* Select the current match and scroll the element into view.
 *
 * The DOM wil change in such a way that after selecting the match, the parent
 * child node count will increase at most by 2 or decrease at most by 1. 
 *
 * The edge cases would be normalizing 3 text nodes and splitting 1 text node,
 * the selected text being at the end of the node; or, no previous selection,
 * with a element node added between 2 text nodes, that is, a text node yielding
 * 3 nodes after the selection.
 *
 * Returns the selection node.
 */
function setFindSelection(node, dict) {
    var matchLength = dict["len"];
    var matchPosition = dict["idx"];
    
    if (0 === "UIWebView would display selections") {
        var selection = window.getSelection();
        selection.removeAllRanges();
    }                        
                        
    parentBeforeSelection = node.parentNode;
    previousSiblingBeforeSelection = node.previousSibling;
                        
    var range = document.createRange();
    range.setStart(node, matchPosition);
    range.setEnd(node, matchPosition + matchLength);
    if (0 === "UIWebView would display selections") {
        selection.addRange(range);
    }

    var selection = document.createElement("span");
    selection.id = "searchResult";
    range.surroundContents(selection);
    
    /* Scroll to the recent selection but don't let it under the toolbar. */
    var top = selection.documentOffsetTop() - (window.innerHeight / 6);
                        
    if (selection.offsetLeft > window.innerWidth) {
        window.scrollTo(selection.offsetLeft, top);
    } else {
        window.scrollTo(0, top);
    }
    
    return selection;
}

function findNextEx(theNode) {
    var results = Array(); 

    switch (theNode.nodeType) {
        case Node.TEXT_NODE:
            /* Find the next match. */
            var thisTime = 0;
            var theText = theNode.textContent; 
            
            /* To compare the term length with the text contents length the 
             * escape characters should not be counted. */
            if (theText.length >= rawTerm.length && theText !== "\n" && theText !== " ")
            {
                var indexes = Array();
                var tmp = re.exec(theText);
                for (; tmp !== null; tmp = re.exec(theText)) {
                    thePosition = { "idx": tmp.index, "len": tmp[0].length, };
                    indexes.push(thePosition);
                    ++thisTime;
                }
            }
            
            if (thisTime > 0) {
                occurrences += thisTime;
                results.push({ "node": theNode, "positions": indexes, });
            }
            break;
        
        case Node.ELEMENT_NODE:
            if ((theNode.className || " ")[0] !== "l") {
                for (var i = 0; i < theNode.childNodes.length; ++i) {
                    results = results.concat(findNextEx(theNode.childNodes.item(i)));
                }
            }
            break;
    }
    
    return results;
}

function findNext() {
    var found = false;
    var status = "true";  
    var normalized = clearFindSelection();
    
    while (queuedMatches.length == 0 && root !== null) {
        var nextTextMatches = findNextEx(root);
        queuedMatches = queuedMatches.concat(nextTextMatches);
        root = root.nextSibling;
    }

    var nextMatch = queuedMatches.shift();
    if (nextMatch !== undefined) {
        found = true;

        var theNode = nextMatch["node"];
        var matches = nextMatch["positions"];
        var theIndex = matches.shift();
        
        /* Setting the selection changes the DOM structure, invalidating the
         * match index(es). However, clearing the selection corrects the offset
         * (though the renormalized node will be different). */
        if (theIndex !== undefined) {
            if (theNode === "denormalized" && normalized !== null) {
                theNode = normalized;
            }
            if (typeof theNode === "object" && "nodeType" in theNode) {
                setFindSelection(theNode, theIndex);
            }
        }
        
        /* If there are any matches left on the text node, restore them to the
         * queued matches, but note that the node was changed. */
        if (matches.length > 0) {
            var remaining = {"node": "denormalized", "positions": matches};
            queuedMatches.splice(0, 0, remaining);
        }
    }
  
    /* Start over when the end of the text is reached. */
    if (queuedMatches.length == 0 && root === null) {
        if (occurrences > 0) {
            resetQueue();
            status = "eof";
        } else {
            /* This is only relevant if the UI is implemented in HTML. */
            var buttonElement = document.getElementById("findNextButton");
            if (buttonElement) {
                buttonElement.setAttribute("disabled");
            }
            status = "false";
        }
    }
    
    return status;
}

/* This is the onchange event handler if the UI is implemented with HTML. */
function searchTermDidChange() {
    var input = document.getElementById("searchInputField");
    var button = document.getElementById("findNextButton");
    var newTerm = input.value.trim();
    input.value = newTerm;
    
    if (setTerm(newTerm)) {
        if (Boolean(newTerm)) {
            button.removeAttribute("disabled");
        } else {     
            button.setAttribute("disabled");
        }
    }
    return false;
}
